#include <stdio.h>
#include <errno.h>
#include <dsassert.h>

#include <dslib.h>
#include <dsmalloc.h>
#include <dsilock.h>

static void dsILockZeroContext(DSILockContext ctx) {
    ctx->initialized = 0;
    ctx->pid = 0;
#if defined(HAVE_PTHREAD_H)
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->lock_mutex = NULL;
    ctx->pid_mutex = NULL;
#endif /* defined(HAVE_PTHREAD_H) */
}

DSILockContext dsILockCreateContext(DSMallocContext mctx) {
    DSILockContext ctx;
    ctx = (DSILockContext)dsMalloc(mctx, sizeof(struct DSILockContextT));
    if (ctx) {
	ctx->malloc = mctx;
	dsILockZeroContext(ctx);
    }
    return ctx;
}

int dsILockInitContext(DSILockContext ctx) {
#if defined(HAVE_PTHREAD_H)
    int err;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_param(ctx);
    dsILockZeroContext(ctx);
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_init(&ctx->lock_mutex,NULL);
# else
    err = pthread_mutex_init(&ctx->lock_mutex,NULL);
# endif /* DS_SUPPORT_RWLOCK */
    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_INIT;
#else /* defined(HAVE_WIN32_THREADS) */
    ctx->lock_mutex = CreateMutex(NULL,FALSE,NULL);
    if (!ctx->lock_mutex) return DS_ERR_WINAPI_MUTEX_CREATE;
#endif /* defined(HAVE_PTHREAD_H) */
    ctx->initialized = DS_INITIALIZED;
    return 0;
}

void dsILockFreeContext(DSILockContext ctx) {
    ds_assert_return(ctx);
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    if (ctx->initialized) pthread_rwlock_destroy(&ctx->lock_mutex);
# else
    if (ctx->initialized) pthread_mutex_destroy(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->lock_mutex) CloseHandle(ctx->lock_mutex);
#endif /* defined(HAVE_PTHREAD_H) */
    dsILockZeroContext(ctx);
}

void dsILockDestroyContext(DSILockContext ctx) {
    ds_assert_return(ctx);
    dsILockFreeContext(ctx);
    dsFree(ctx->malloc, ctx);
}

int dsILockRead(DSILockContext ctx) {
    int err;
    ds_assert_ctx(ctx);
#if defined(HAVE_PTHREAD_H)
    if (pthread_equal(ctx->pid,pthread_self())) return DS_ERR_LOCK_SELFLOCKED;
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_rdlock(&ctx->lock_mutex);
# else
    err = pthread_mutex_lock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->pid == GetCurrentThreadId()) return DS_ERR_LOCK_SELFLOCKED;
    if (WaitForSingleObject(ctx->lock_mutex, INFINITE) == WAIT_OBJECT_0) err = 0;
    else err = -1;
#endif /* defined(HAVE_PTHREAD_H) */

    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    return 0;
}

int dsILockWrite(DSILockContext ctx) {
    int err;
    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
    if (pthread_equal(ctx->pid,pthread_self())) return DS_ERR_LOCK_SELFLOCKED;
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_wrlock(&ctx->lock_mutex);
# else
    err = pthread_mutex_lock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
    if (!err) ctx->pid = pthread_self();
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->pid == GetCurrentThreadId()) return DS_ERR_LOCK_SELFLOCKED;
    if (WaitForSingleObject(ctx->lock_mutex, INFINITE) == WAIT_OBJECT_0) {
	err = 0;
	ctx->pid = GetCurrentThreadId();
    } else err = -1;
#endif /* defined(HAVE_PTHREAD_H) */

    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    return 0;
}

int dsILockTryRead(DSILockContext ctx) {
#if defined(HAVE_PTHREAD_H)
    int err;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
    if (pthread_equal(ctx->pid,pthread_self())) return DS_ERR_LOCK_SELFLOCKED;
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_tryrdlock(&ctx->lock_mutex);
# else
    err = pthread_mutex_trylock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
    if (err) {
	if (err == EDEADLK) return DS_ERR_LOCK_LOCKED;
	return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    }
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->pid == GetCurrentThreadId()) return DS_ERR_LOCK_SELFLOCKED;
    if (WaitForSingleObject(ctx->lock_mutex, 0) != WAIT_OBJECT_0) return DS_ERR_LOCK_LOCKED;
#endif /* defined(HAVE_PTHREAD_H) */

    return 0;
}

int dsILockTryWrite(DSILockContext ctx) {
#if defined(HAVE_PTHREAD_H)
    int err;
#endif /* defined(HAVE_PTHREAD_H) */

    ds_assert_ctx(ctx);

#if defined(HAVE_PTHREAD_H)
    if (pthread_equal(ctx->pid,pthread_self())) return DS_ERR_LOCK_SELFLOCKED;
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_trywrlock(&ctx->lock_mutex);
# else
    err = pthread_mutex_trylock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
    if (err) {
	if (err == EDEADLK) return DS_ERR_LOCK_LOCKED;
	return DS_ERR_SYSCALL_PTHREAD_MUTEX_LOCK;
    } else ctx->pid = pthread_self();
#else /* defined(HAVE_WIN32_THREADS) */
    if (ctx->pid == GetCurrentThreadId()) return DS_ERR_LOCK_SELFLOCKED;
    if (WaitForSingleObject(ctx->lock_mutex, 0) == WAIT_OBJECT_0) ctx->pid = GetCurrentThreadId();
    else return DS_ERR_LOCK_LOCKED;
#endif /* defined(HAVE_PTHREAD_H) */
    return 0;
}

int dsILockReturn(DSILockContext ctx) {
    int err;
    ds_assert_ctx(ctx);

    ctx->pid = 0;
#if defined(HAVE_PTHREAD_H)
# ifdef DS_SUPPORT_RWLOCK
    err = pthread_rwlock_unlock(&ctx->lock_mutex);
# else
    err = pthread_mutex_unlock(&ctx->lock_mutex);
# endif /* DS_SUPPORT_RWLOCK */
#else /* defined(HAVE_WIN32_THREADS) */
    if (ReleaseMutex(ctx->lock_mutex)) err = 0;
    else err = -1;
#endif /* defined(HAVE_PTHREAD_H) */

    if (err) return DS_ERR_SYSCALL_PTHREAD_MUTEX_UNLOCK;
    return 0;
}
