#define _OPC_DATA_C

#include <opclib.h>
#include <dsassert.h>

#include <dserror.h>
#include <dsdebug.h>
#include <opcdata.h>

#define OPC_DATA_ALLOCATE_PRESENTATIONS 2


OPCDataOptions opcDataGetDefaultOptions() {
    int err;
    OPCDataOptions options;

    options = (OPCDataOptions)malloc(sizeof(struct OPCDataOptionsT));
    if (!options) return options;
    
    err = opcDataInitDefaultOptions(options);
    if (err) {
	opcDataDestroyOptions(options);
	return NULL;
    }
    return options;
}

int opcDataInitDefaultOptions(OPCDataOptions options) {
    ds_assert_param(options);
    memcpy(options, &opc_datamanager_data_options, sizeof(struct OPCDataOptionsT));
    return 0;
}

void opcDataDestroyOptions(OPCDataOptions options) {
    if (options) free(options);
}


int opcDataOptionsSetAcquireFunction(OPCDataOptions options, OPCDataManagerAcquireFunction acquire) {
    ds_assert_param(options);
    options->acquire = acquire;
    return 0;
}

int opcDataOptionsSetBufferSize(OPCDataOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->binary_minbuf_size = min;
    options->binary_maxbuf_size = max;
    return 0;
}

int opcDataOptionsSetCacheSize(OPCDataOptions options, unsigned long min, unsigned long max) {
    ds_assert_param(options);
    options->cache_min = min;
    options->cache_max = max;
    return 0;
}

int opcDataOptionsSetFormatsContext(OPCDataOptions options, OPCFormatsContext formats) {
    ds_assert_param(options);
    options->formats = formats;
    return 0;
}

int opcDataOptionsSetBinaryAllocator(OPCDataOptions options, DSAllocatorContext allocator) {
    ds_assert_param(options);
    options->binary_allocator = allocator;
    return 0;
}

int opcDataOptionsSetBufferPool(OPCDataOptions options, DSBufferPoolContext buffer_pool) {
    ds_assert_param(options);
    options->buffer_pool = buffer_pool;
    return 0;
}

int opcDataOptionsSetChunkPool(OPCDataOptions options, OPCChunkPoolContext chunk_pool) {
    ds_assert_param(options);
    options->chunk_pool = chunk_pool;
    return 0;
}

int opcDataOptionsSetDataSourceFlags(OPCDataOptions options, unsigned long flags) {
    ds_assert_param(options);
    options->flags = flags;
    return 0;
}


void opcDataZeroContext(OPCDataContext ctx) {
    ds_assert_return(ctx);

    ctx->representation = NULL;
    ctx->repinfo = NULL;
    ctx->rephash = NULL;
    ctx->native_representation = NULL;
    ctx->opcxml_representation = NULL;
    
    ctx->chunks = NULL;
    ctx->current = NULL;
    ctx->lock = NULL;
    ctx->representation_lock = NULL;
    ctx->chunk_lock = NULL;
    
    ctx->nchunks = 0;
}

static void opcDataFreeContext(OPCDataContext ctx) {
    OPCDataChunkContext chunk, nchunk;

    dsDebugServer(DS_DEBUG_NOTICE, "opcDataFreeContext: %p", ctx);    

    ds_assert_return(ctx);
    
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    if (ctx->chunk_lock) dsLockDestroyContext(ctx->chunk_lock);
    if (ctx->representation_lock) dsLockDestroyContext(ctx->representation_lock);

    for (chunk=ctx->chunks;chunk;chunk=nchunk) {
        nchunk = chunk->next;
	opcDataChunkWaitUnref(chunk, 0);
        opcDataChunkDestroyContext(chunk);
    }

    if (ctx->rephash) dsHashTableDestroy(ctx->rephash);    
    if (ctx->representation) dsFree(ctx->malloc, ctx->representation);
    if (ctx->repinfo) dsFree(ctx->malloc, ctx->repinfo);
    
    opcDataZeroContext(ctx);    
}

static void opcDataDestroyContext(OPCDataContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcDataDestroyContext: %p", ctx);    
    opcDataFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}


OPCDataContext opcDataCreateContext(DSMallocContext mctx) {
    OPCDataContext ctx;

    ctx = (OPCDataContext)dsMalloc(mctx,sizeof(struct OPCDataContextT));
    dsDebugServer(DS_DEBUG_NOTICE, "opcDataCreateContext: %p", ctx);    
    if (ctx) {
	ctx->malloc=mctx;
	ctx->refs = 1;
	opcDataZeroContext(ctx);
	memcpy(&ctx->manager, &opc_datamanager_data, sizeof(struct OPCDataManagerContextT));
    }
    return ctx;
}

int opcDataInitContext(OPCDataContext ctx, OPCDataOptions options) {
    int err;
    unsigned short int i;

    dsDebugServer(DS_DEBUG_NOTICE, "opcDataInitContext: %p", ctx);

    ds_assert_param(ctx);
    ds_assert_param(options);
    ds_assert_options(options->chunk_pool);
    
    ctx->manager.Acquire = options->acquire;
    ctx->formats = options->formats;
    ctx->binary_allocator = options->binary_allocator;
    ctx->buffer_pool = options->buffer_pool;
    ctx->chunk_pool = options->chunk_pool;
    
    if (options->cache_min) ctx->cache_min = options->cache_min;
    else ctx->cache_min = 1;
    ctx->cache_max = options->cache_max;
    
    ctx->binary_minbuf_size = options->binary_minbuf_size;
    ctx->binary_maxbuf_size = options->binary_maxbuf_size;
    
    ctx->flags = options->flags;
    
    opcDataZeroContext(ctx);    

    ctx->allocated_representations = OPC_DATA_ALLOCATE_PRESENTATIONS;

    ctx->representation = (OPCRepresentationContext*)dsMalloc(ctx->malloc, ctx->allocated_representations*sizeof(OPCDataRepresentationContext));
    ctx->repinfo = (OPCRepresentationInfo*)dsMalloc(ctx->malloc, ctx->allocated_representations*sizeof(OPCRepresentationInfo));
    ctx->rephash = dsHashTableInit(ctx->malloc, OPC_DATA_HASHTABLE_SIZE, DS_HASH_TYPE_STATICPOINTER, NULL);
    if ((!ctx->representation)||(!ctx->repinfo)||(!ctx->rephash)) {
	opcDataFreeContext(ctx);
	return DS_ERR_MEM;
    }
    dsHashTableSetDestructor(ctx->rephash, (HashTableDestructorFunction)opcRepresentationUnref);


    ctx->supported_representations = 2;
    for (i=0;i<ctx->supported_representations;i++)
	ctx->representation[i] = NULL;
    
    ctx->lock = dsLockCreateContext(ctx->malloc);
    ctx->chunk_lock = dsLockCreateContext(ctx->malloc);
    ctx->representation_lock = dsLockCreateContext(ctx->malloc);
    if ((!ctx->lock)||(!ctx->chunk_lock)||(!ctx->representation_lock)) {
	opcDataFreeContext(ctx);
	return DS_ERR_LOCK_CREATE;
    }
    err = dsLockInitContext(ctx->lock);
    if (!err) err = dsLockInitContext(ctx->chunk_lock);
    if (!err) err = dsLockInitContext(ctx->representation_lock);
    if (err) {
	opcDataFreeContext(ctx);
	return err;
    }

    return 0;
}


int opcDataRef(OPCDataContext ctx) {
    ds_assert_param(ctx);
    
    dsLockWrite(ctx->lock);
    ctx->refs++;
    dsLockReturn(ctx->lock);
    return 0;
}

void opcDataUnref(OPCDataContext ctx) {
    ds_assert_return(ctx);
    if (!ctx->lock) {
	opcDataDestroyContext(ctx);
	return;
    }
    
    dsLockWrite(ctx->lock);
    if ((--ctx->refs)==0) {
	dsLockReturn(ctx->lock);
	opcDataDestroyContext(ctx);
	return;
    }
    dsLockReturn(ctx->lock);
}


int opcDataSetDataRepresentationOptions(OPCDataContext ctx, OPCDataRepresentationOptions options, OPCFormatContext format) {
    ds_assert_param(ctx);
    ds_assert_param(options);
    ds_assert_param(format);
    
    if ((!ctx)||(!options)||(!format)) return DS_ERR_INVALID_ARGUMENT;
    
    opcDataRepresentationOptionsSetFormat(options, format);
    opcDataRepresentationOptionsSetBufferSize(options, ctx->binary_minbuf_size, ctx->binary_maxbuf_size);
    opcDataRepresentationOptionsSetDefaultAllocatorOptions(options,ctx->buffer_pool);
    opcDataRepresentationOptionsSetAllocator(options, ctx->binary_allocator);
    opcDataRepresentationOptionsSetFlags(options, ctx->flags);
    return 0;
}

unsigned long opcDataGetId(OPCDataContext ctx, const char *repname) {
    OPCRepresentationContext rep;
    
    if ((!ctx)||(!repname)||(!ctx->rephash)) return (unsigned long)-1;
    
    dsLockRead(ctx->representation_lock);
    rep=dsHashTableFind(ctx->rephash, (void*)repname);
    dsLockReturn(ctx->representation_lock);
    
    if (!rep) return (unsigned long)-1;
    return rep->id;
}


static int opcDataRegisterRepresentationI(OPCDataContext ctx, OPCRepresentationContext rep, unsigned short int *repnum) {
    OPCRepresentationContext *representations;
    unsigned short int repid;
    
    repid = *repnum;

    if (!rep->title) return OPC_ERR_REPRESENTATION_INCOMPLETE;
    
    dsLockWrite(ctx->representation_lock);
    if (repid==(unsigned short int)-1) {
	if (ctx->supported_representations==ctx->allocated_representations) {
	    representations = (OPCRepresentationContext*)dsRealloc(ctx->malloc, ctx->representation, 2*ctx->allocated_representations*sizeof(OPCDataRepresentationContext));
	    if (!representations) return DS_ERR_MEM;
	    ctx->representation = representations;
	    ctx->allocated_representations*=2;
	}
	repid = ctx->supported_representations;	/* Isn't default representation and can't be accessed until in hash */
    } else {
        if (ctx->representation[repid]) return OPC_ERR_DATAMANAGER_REPRESENTATION_ALREADY_REGISTERED;
    }

    opcRepresentationSetId(rep, repid);
    if (dsHashTableAdd(ctx->rephash, (void*)rep, 1)) return DS_ERR_MEM;
    ctx->representation[repid] = rep;
    ctx->repinfo[repid] = &rep->repinfo;
    ctx->supported_representations++;
    if (*repnum == (unsigned short int)-1) *repnum = repid;
    dsLockReturn(ctx->representation_lock);

    dsDebugServer(DS_DEBUG_DEBUG, "opcData(%p): AddRepresentation(%s): Registered as %u.", ctx, rep->title, repid);

    return 0;
}

unsigned short int opcDataRegisterRepresentation(OPCDataContext ctx, OPCRepresentationContext representation) {
    int err;
    unsigned short int rep = (unsigned short int)-1;
    
    err = opcDataRegisterRepresentationI(ctx, representation, &rep);
    if (err) {
	dsDebugServer(DS_DEBUG_WARNING, "opcData(%p): AddRepresentation(%p): Failed: %s.", ctx, representation, dsError(err, NULL));
	return (unsigned short int)-1;
    }
    
    return rep;
}

int opcDataRegisterNativeRepresentation(OPCDataContext ctx, OPCType type, OPCDataRepresentationOptions options) {
    int err;
    OPCDataRepresentationOptions opts;
    OPCDataRepresentationContext rep;
    OPCFormatsContext formats;
    unsigned short int repid = OPC_REPRESENTATION_NATIVE;

    if ((!ctx)||(!type)) return DS_ERR_INVALID_ARGUMENT;

    formats = ctx->formats;
    if ((!formats)||(!formats->native_format)) return OPC_ERR_FORMATS_INCOMPLETE;
    

    rep = opcDataRepresentationCreateContext(ctx->malloc, "Native");
    if (!rep) return OPC_ERR_REPRESENTATION_CREATE_NATIVE;
    
    if (options) opts=options;
    else {
	opts = opcDataRepresentationGetDefaultOptions();
	opcDataSetDataRepresentationOptions(ctx, opts, (OPCFormatContext)formats->native_format);
	opcDataRepresentationOptionsSetType(opts, type);
	opcDataRepresentationOptionsSetBase(opts, NULL);
    }

    err = opcDataRepresentationInitContext(rep, opts);
    if (err) opcDataRepresentationUnref(rep);
    else {
	err = opcDataRegisterRepresentationI(ctx, (OPCRepresentationContext)rep, &repid);
	if (err) opcDataRepresentationUnref(rep);
	else ctx->native_representation = rep;
    }

    if (!options) opcDataRepresentationDestroyOptions(opts);
    return err;
}

int opcDataRegisterOPCXMLRepresentation(OPCDataContext ctx, OPCType type, OPCDataRepresentationOptions options) {
    int err;
    OPCDataRepresentationOptions opts;
    OPCDataRepresentationContext rep;
    OPCFormatsContext formats;
    unsigned short int repid = OPC_REPRESENTATION_OPCXML;
    
    ds_assert_param(ctx);
    ds_assert_param(type);
    
    ds_assert_err(ctx->native_representation, OPC_ERR_REPRESENTATION_MISSING_NATIVE);

    formats = ctx->formats;
    ds_assert_err(formats, OPC_ERR_FORMATS_MISSING);
    ds_assert_err(formats->opcxml_format, OPC_ERR_FORMATS_MISSING_OPCXML);

    rep = opcDataRepresentationCreateContext(ctx->malloc, "OPCXML");
    if (!rep) return OPC_ERR_REPRESENTATION_CREATE_OPCXML;
    
    if (options) opts=options;
    else {
	opts = opcDataRepresentationGetDefaultOptions();
	opcDataSetDataRepresentationOptions(ctx, opts, formats->opcxml_format);
	opcDataRepresentationOptionsSetType(opts, type);
	opcDataRepresentationOptionsSetBase(opts, (OPCRepresentationContext)ctx->native_representation);
    }

    err = opcDataRepresentationInitContext(rep, opts);
    if (err) opcDataRepresentationUnref(rep);
    else {
	err = opcDataRegisterRepresentationI(ctx, (OPCRepresentationContext)rep, &repid);
	if (err) opcDataRepresentationUnref(rep);
	else ctx->opcxml_representation = rep;
    }

    if (!options) opcDataRepresentationDestroyOptions(opts);

    return 0;
}


static inline_function void opcDataClear(OPCDataContext ctx) {
    DSLockContext chunk_lock;
    OPCChunkPoolContext chunk_pool;
    OPCDataChunkContext chunk, nchunk;
    unsigned long buffer_min, buffer_max;
    unsigned long ref = 0;

	/* After chunk_lock locked number of referers can decrease, but not increase! */

    chunk_lock = ctx->chunk_lock;
    chunk_pool = ctx->chunk_pool;
    chunk = ctx->chunks;

    buffer_min = ctx->cache_min;
    buffer_max = ctx->cache_max;

    nchunk = chunk;

    if (buffer_max) 
        for (;((ctx->nchunks>buffer_max)&&(!ref));chunk=nchunk) {
            nchunk = chunk->next;
	    ctx->chunks=nchunk;
	    ctx->nchunks--;
	    dsLockReturn(chunk_lock);
	    opcChunkPoolReturnChunk(chunk_pool, chunk);
	    dsLockWrite(chunk_lock);
	    ref=opcDataChunkGetRefs(nchunk);
	}
	
    for (;((!ref)&&(ctx->nchunks>buffer_min));chunk=nchunk) {
	nchunk=chunk->next;
	ctx->chunks=nchunk;
	ctx->nchunks--;
	dsLockReturn(chunk_lock);
	opcChunkPoolReturnChunk(chunk_pool, chunk);
	dsLockWrite(chunk_lock);
	ref = opcDataChunkGetRefs(nchunk);
	if (!ref) ref = opcDataChunkGetDelayedRefs(nchunk);
    }
    
}

DSAllocatorContext opcDataGetAllocator(OPCDataContext ctx, unsigned long repid, OPCDataInfo info) {
    OPCRepresentationContext rctx;

    if ((!ctx)||(repid>=ctx->supported_representations)) return NULL;

    dsLockRead(ctx->representation_lock);
    ds_assert_null(ctx->representation);
    rctx = ctx->representation[repid];
    dsLockReturn(ctx->representation_lock);
    ds_assert_null(rctx);
    
    if (!rctx) return NULL;

    return opcRepresentationGetAllocator(rctx, info);
}

void opcDataReturnAllocator(OPCDataContext ctx, unsigned long repid, DSAllocatorContext allocator) {
    OPCRepresentationContext rctx;

    if ((!ctx)||(repid>=ctx->supported_representations)) return;

    dsLockRead(ctx->representation_lock);
    ds_assert_info(ctx->representation);
    if (ctx->representation) rctx = ctx->representation[repid];
    else rctx = NULL;
    dsLockReturn(ctx->representation_lock);
    ds_assert_return(rctx);
    
    if (!rctx) return;
    
    opcRepresentationReturnAllocator(rctx, allocator);
}

int opcDataPushAllocation(OPCDataContext ctx, unsigned long repid, DSAllocatorContext allocator, OPCDataInfo info, DSAllocation allocation) {
    int err;
    OPCRepresentationContext rctx;
    OPCDataChunkContext chunk;

    ds_assert_param(ctx);
    ds_assert_param(allocation);
    ds_assert_param(repid>=ctx->supported_representations);

    dsLockRead(ctx->representation_lock);
    ds_assert_param(ctx->representation);
    rctx = ctx->representation[repid];
    dsLockReturn(ctx->representation_lock);
    ds_assert_param(rctx);

    if (!rctx) return OPC_ERR_REPRESENTATION_DISABLED;

    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, ctx->repinfo, ctx->supported_representations);
    if (!chunk) return OPC_ERR_CHUNKPOOL_GET;
    opcDataChunkCopyInfo(chunk, info);

    dsLockWrite(ctx->lock);
    err = opcRepresentationPushAllocation(rctx, chunk, allocator, allocation);
	// ToDo: prelock datachunk for all subscriptions may need it
    dsLockReturn(ctx->lock);

    if (err) {
	opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
	return err;
    }
    
    dsLockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (dsTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current=chunk;

	if ((ctx->nchunks>ctx->cache_min)&&(!opcDataChunkGetRefs(ctx->chunks))) 
	    opcDataClear(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    dsLockReturn(ctx->chunk_lock);
    

    return 0;
    
}



int opcDataPushData(OPCDataContext ctx, unsigned long repid, OPCDataInfo info, const void *data, unsigned long size) {
    int err;
    OPCRepresentationContext rctx;
    OPCDataChunkContext chunk;

    ds_assert_param(ctx);
    ds_assert_param(data);
    ds_assert_param(repid<ctx->supported_representations);

    dsLockRead(ctx->representation_lock);
    ds_assert_param(ctx->representation);
    rctx = ctx->representation[repid];
    dsLockReturn(ctx->representation_lock);
    
    if (!rctx) return OPC_ERR_REPRESENTATION_DISABLED;

    dsDebugServer(DS_DEBUG_DEBUG, "opcData(%p): Push (representation: %u).", ctx, repid);
    
    chunk = opcChunkPoolGetChunk(ctx->chunk_pool, ctx->repinfo, ctx->supported_representations);
    if (!chunk) return OPC_ERR_CHUNKPOOL_GET;
    opcDataChunkCopyInfo(chunk, info);

    dsLockWrite(ctx->lock);
    err = opcRepresentationPushData(rctx, chunk, data, size);
	// ToDo: prelock datachunk for all subscriptions may need it
    dsLockReturn(ctx->lock);

    if (err) {
	opcChunkPoolReturnChunk(ctx->chunk_pool, chunk);
	return err;
    }
    
    dsLockWrite(ctx->chunk_lock);
    ctx->nchunks++;
    if (ctx->current) {
	/* What to do with OOB items, with time earlier than current one?
	For Fast items can be assumed only one sequential data source? For items
	with multiple or/and non-sequential data sources OPCDataContext should be used
	if (dsTimeCmp(ctx->current->info.time, info->time)>0) {
	}
	*/
	ctx->current->next = chunk;
	ctx->current=chunk;

	if ((ctx->nchunks>ctx->cache_min)&&(!opcDataChunkGetRefs(ctx->chunks))) 
	    opcDataClear(ctx);
    } else {
	ctx->chunks = chunk;
	ctx->current = chunk;
    }
    dsLockReturn(ctx->chunk_lock);
    

    return 0;
}

int opcDataGetData(OPCDataContext ctx, unsigned long repid, OPCDataInfo *info, OPCRepresentationData *data) {
    int err;
    OPCRepresentationContext rep;
    OPCDataChunkContext current;

    ds_assert_complexctx(ctx, (ctx->current)&&(ctx->representation));
    ds_assert_param(repid<ctx->supported_representations);

    dsLockRead(ctx->chunk_lock);
    current = ctx->current;
    opcDataChunkRef(current);
    dsLockReturn(ctx->chunk_lock);
    
    if (data) {    
	dsLockRead(ctx->representation_lock);
	rep = ctx->representation[repid];
	if (rep) err = opcRepresentationGetData(rep, current, data);
	else err = OPC_ERR_REPRESENTATION_DISABLED;
	dsLockReturn(ctx->representation_lock);
    }

    if (!err) {
	if (info) *info=opcDataChunkGetInfo(current);
        opcDataRef(ctx);
    }
    
    return err;
}


void opcDataReturnData(OPCDataContext ctx, OPCDataInfo info) {
    ds_assert_return(ctx);
    if ((!ctx)||(!info)) return;

    dsLockWrite(ctx->chunk_lock);
    if ((!opcDataChunkUnrefByInfo(info))&&(ctx->chunks == opcDataChunkGetChunkByInfo(info)))
	opcDataClear(ctx);
    dsLockReturn(ctx->chunk_lock);
    
    opcDataUnref(ctx);
}

void opcDataMantain(OPCDataContext ctx) {
    OPCRepresentationContext *representation;
    long i;

    ds_assert_return(ctx);
    if ((!ctx)||(!ctx->representation)) return;
    
    dsLockRead(ctx->representation_lock);
    representation = ctx->representation;
    for (i=ctx->supported_representations;i>=0;i--) 
	if (representation[i]) opcRepresentationMantain(representation[i]);
    dsLockReturn(ctx->representation_lock);
}
