#include <dsassert.h>

#include <dslib.h>
#include <dsdebug.h>
#include <dserror.h>
#include <opcchunkpool.h>

void opcChunkPoolZeroContext(OPCChunkPoolContext ctx) {
    ds_assert_return(ctx);
    ctx->first_chunk = NULL;
    ctx->last_chunk = NULL;
    ctx->lock = NULL;
}

OPCChunkPoolContext opcChunkPoolCreateContext(DSMallocContext mctx) {
    OPCChunkPoolContext ctx;

    dsDebugServer(DS_DEBUG_NOTICE, "opcChunkPoolCreateContext");    

    ctx = (OPCChunkPoolContext)dsMalloc(mctx,sizeof(struct OPCChunkPoolContextT));
    if (ctx) {
	ctx->malloc=mctx;
	opcChunkPoolZeroContext(ctx);
    }
    return ctx;
}

int opcChunkPoolInitContext(OPCChunkPoolContext ctx, unsigned char bins) {
    int err;
    unsigned char i;
    OPCDataChunkContext *ptr1, *ptr2;

    dsDebugServer(DS_DEBUG_NOTICE, "opcChunkPoolInitContext");    

    ds_assert_param(ctx);
    ds_assert_param(bins);

    ptr1 = (OPCDataChunkContext*)dsMalloc(ctx->malloc, bins*sizeof(OPCDataChunkContext));
    ptr2 = (OPCDataChunkContext*)dsMalloc(ctx->malloc, bins*sizeof(OPCDataChunkContext));
    if ((!ptr1)||(!ptr2)) {
	if (ptr1) dsFree(ctx->malloc, ptr1);
	if (ptr2) dsFree(ctx->malloc, ptr2);
	return DS_ERR_MEM;
    }
    
    for (i=0;i<bins;i++) {
	ptr1[i]=NULL;
	ptr2[i]=NULL;
    }

    ctx->lock = dsLockCreateContext(ctx->malloc);
    if (!ctx->lock) {
	opcChunkPoolFreeContext(ctx);
	return DS_ERR_MEM;
    }
    err = dsLockInitContext(ctx->lock);
    if (err) {
	opcChunkPoolFreeContext(ctx);
	return err;
    }

    ctx->bins = bins;
    ctx->first_chunk = ptr1;
    ctx->last_chunk = ptr2;
    
    return 0;
}

void opcChunkPoolFreeContext(OPCChunkPoolContext ctx) {
    unsigned char i;
    OPCDataChunkContext ptr1, ptr2;

    dsDebugServer(DS_DEBUG_NOTICE, "opcChunkPoolFreeContext");    

    ds_assert_return(ctx);
    
    if (ctx->lock) dsLockDestroyContext(ctx->lock);
    
    if (ctx->first_chunk) {
	for (i=0;i<ctx->bins;i++) {
	    for (ptr1=ctx->first_chunk[i];ptr1;ptr1=ptr2) {
		ptr2 = ptr1->next;
		opcDataChunkDestroyContext(ptr1);
	    }
	}
	dsFree(ctx->malloc, ctx->first_chunk);
    }
    if (ctx->last_chunk) dsFree(ctx->malloc, ctx->last_chunk);

    opcChunkPoolZeroContext(ctx);    
}

void opcChunkPoolDestroyContext(OPCChunkPoolContext ctx) {
    dsDebugServer(DS_DEBUG_NOTICE, "opcChunkPoolDestroyContext");    
    opcChunkPoolFreeContext(ctx);
    dsFree(ctx->malloc,ctx);
}

OPCDataChunkContext opcChunkPoolGetChunk(OPCChunkPoolContext ctx, OPCRepresentationInfo *definfo, unsigned short int formats) {
    int err;
    OPCDataChunkContext chunk;
    unsigned short int i;
    unsigned char j;
    
    ds_assert_null(ctx);
    ds_assert_null(formats);

    for (j=0,i=(formats>>1);i;i>>=1) j++;
    if (formats != (1<<j)) j++;
    i = (1<<j);
    
    if (j>=ctx->bins) {
	dsDebugServer(DS_DEBUG_WARNING, "opcChunkPoolGetChunk: You should increase number of bins.");
	j=ctx->bins-1;
    }

    
    dsLockWrite(ctx->lock);
    
    chunk = ctx->first_chunk[j];
    if (chunk) {
	if (chunk->next) {
	    ctx->first_chunk[j]=chunk->next;
	} else {
	    ctx->first_chunk[j]=NULL;
	    ctx->last_chunk[j]=NULL;
	}
	opcDataChunkRestoreContext(chunk, definfo);
    } else {
	chunk = opcDataChunkCreateContext(ctx->malloc);
	if (chunk) {
	    err = opcDataChunkInitContext(chunk, definfo, i);
	    if (err) {
		dsDebugServer(DS_DEBUG_ERROR, "opcDataChunkInitContext: %lx (%s)", err, dsError(err, NULL));
		opcDataChunkDestroyContext(chunk);
		chunk = NULL;
	    }
	}
    }
    dsLockReturn(ctx->lock);

    return chunk;
}

void opcChunkPoolReturnChunk(OPCChunkPoolContext ctx, OPCDataChunkContext chunk) {
    unsigned short int i;
    unsigned char j;

    ds_assert_return(ctx);
    ds_assert_return(chunk);

    opcDataChunkFreeContext(chunk);
    
    for (j=0,i=(chunk->supported_representations>>1);i;i>>=1) j++;
    if (chunk->supported_representations != (1<<j)) j++;
    
    if (j>=ctx->bins) {
	dsDebugServer(DS_DEBUG_WARNING, "opcChunkPoolReturnChunk: You should increase number of bins.");
	j=ctx->bins-1;
    }

    dsLockWrite(ctx->lock);
    if (ctx->last_chunk[j]) {
	ctx->last_chunk[j]->next = chunk;
	ctx->last_chunk[j] = chunk;
    } else {
	ctx->first_chunk[j] = chunk;
	ctx->last_chunk[j] = chunk;
    }
    chunk->next = NULL;    
    dsLockReturn(ctx->lock);
}

